@@ -10,7 +10,7 @@ from pydantic import Field, HttpUrl, SecretStr
 from pydantic_settings import BaseSettings, SettingsConfigDict
 
 from moxn.exceptions import MoxnSchemaValidationError
-from moxn.models.prompt import Prompt, PromptInstance
+from moxn.models.prompt import Prompt, PromptInstance, RenderableModel
 from moxn.models.schema import PromptSchemas, SchemaPromptType, CodegenResponse
 from moxn.models.task import Task
 from moxn.polling import PollingConfig, PollingManager
@@ -20,6 +20,8 @@ from moxn.base_models.telemetry import (
     SpanLogRequest,
     TelemetryLogResponse,
     LLMEvent,
+    SignedUrlRequest,
+    SignedUrlResponse,
 )
 from contextlib import asynccontextmanager
 from typing import Any, AsyncGenerator, Literal, Optional
@@ -36,14 +38,16 @@ class MoxnSettings(BaseSettings):
     user_id: str
     org_id: str | None = None
     api_key: SecretStr
-    base_api_route: HttpUrl = Field(default=HttpUrl("https://api.moxn.io/v1"))
+    base_api_route: HttpUrl = Field(
+        default=HttpUrl("https://marksweissma--moxn-api-fastapi-app.modal.run")
+    )
     timeout: float = Field(default=60.0, description="Prompt timeout in seconds")
 
     model_config = SettingsConfigDict(
         frozen=True,
         extra="forbid",
+        env_file=[".env.local", ".env"],
         env_prefix="MOXN_",
-        env_file="src/moxn/.env",
         env_file_encoding="utf-8",
     )
 
@@ -107,10 +111,10 @@ class MoxnClient(TelemetryTransportBackend):
         """Returns the default headers for API prompts."""
         headers = {
             "x-api-key": self.settings.api_key.get_secret_value(),
-            "x-prompted-user-id": self.settings.user_id,
+            "x-requested-user-id": self.settings.user_id,
         }
         if self.settings.org_id:
-            headers["x-prompted-org-id"] = self.settings.org_id
+            headers["x-requested-org-id"] = self.settings.org_id
         return headers
 
     async def __aenter__(self) -> "MoxnClient":
@@ -248,6 +252,12 @@ class MoxnClient(TelemetryTransportBackend):
 
         try:
             schemas = response.json()
+            with open("temp.json", "w") as f:
+                f.write(json.dumps(schemas, indent=2))
+
+            # Pre-process schema data to handle self-references
+            self._preprocess_schema_data(schemas)
+
             return PromptSchemas.model_validate(schemas)
         except Exception as e:
             logger.error(
@@ -261,6 +271,70 @@ class MoxnClient(TelemetryTransportBackend):
                 detail=str(e),
             )
 
+    def _preprocess_schema_data(self, schema_data: dict) -> None:
+        """
+        Pre-process schema data to handle self-references and other validation issues.
+
+        This method modifies the schema_data in place to ensure it can be properly
+        validated by the Pydantic model.
+        """
+        # Process inputs
+        if "inputs" in schema_data:
+            for input_schema in schema_data["inputs"]:
+                self._process_schema_properties(
+                    input_schema.get("schema", {}).get("properties", [])
+                )
+
+        # Process outputs
+        if "outputs" in schema_data:
+            for output_schema in schema_data["outputs"]:
+                self._process_schema_properties(
+                    output_schema.get("schema", {}).get("properties", [])
+                )
+
+    def _process_schema_properties(self, properties: list) -> None:
+        """
+        Recursively process schema properties to fix validation issues.
+        """
+        if not properties:
+            return
+
+        for prop in properties:
+            # Handle items property which might contain nested schema references
+            if "items" in prop and isinstance(prop["items"], dict):
+                # Process schemaRef if it exists
+                if "schemaRef" in prop["items"] and isinstance(
+                    prop["items"]["schemaRef"], dict
+                ):
+                    schema_ref = prop["items"]["schemaRef"]
+                    # Convert boolean isSelfReference to string to match expected type
+                    if "isSelfReference" in schema_ref and isinstance(
+                        schema_ref["isSelfReference"], bool
+                    ):
+                        schema_ref["isSelfReference"] = str(
+                            schema_ref["isSelfReference"]
+                        ).lower()
+
+                # Recursively process nested properties
+                if "properties" in prop["items"] and isinstance(
+                    prop["items"]["properties"], list
+                ):
+                    self._process_schema_properties(prop["items"]["properties"])
+
+            # Process nested properties
+            if "properties" in prop and isinstance(prop["properties"], list):
+                self._process_schema_properties(prop["properties"])
+
+            # Handle schemaRef at the property level
+            if "schemaRef" in prop and isinstance(prop["schemaRef"], dict):
+                schema_ref = prop["schemaRef"]
+                if "isSelfReference" in schema_ref and isinstance(
+                    schema_ref["isSelfReference"], bool
+                ):
+                    schema_ref["isSelfReference"] = str(
+                        schema_ref["isSelfReference"]
+                    ).lower()
+
     async def get_code_stubs(
         self,
         prompt_id: str,
@@ -376,14 +450,14 @@ class MoxnClient(TelemetryTransportBackend):
 
     async def create_telemetry_log(
         self,
-        prompt: SpanLogRequest | SpanEventLogRequest,
+        log_request: SpanLogRequest | SpanEventLogRequest,
     ) -> TelemetryLogResponse:
         """Send telemetry log to the API."""
-        logger.debug(f"Sending telemetry log: {prompt}")
+        logger.debug(f"Sending telemetry log: {log_request}")
 
         try:
             # Ensure proper serialization of all fields
-            json_data = prompt.model_dump(
+            json_data = log_request.model_dump(
                 exclude_none=True,
                 mode="json",
                 by_alias=True,
@@ -408,8 +482,7 @@ class MoxnClient(TelemetryTransportBackend):
         self,
         prompt_id: str,
         version_id: str | None = None,
-        message_names: list[str] | None = None,
-        **variables,
+        input_schema: RenderableModel | None = None,
     ) -> PromptInstance:
         """
         Create a new PromptInstance for managing LLM interactions.
@@ -417,33 +490,48 @@ class MoxnClient(TelemetryTransportBackend):
         Args:
             prompt_id: The base prompt ID
             version_id: Optional specific version
-            message_names: Optional list of message names to include
-            **variables: Variables to use in message rendering
+            input_schema: Optional input schema to use in message rendering
         """
         prompt = await self.get_prompt(prompt_id, version_id)
-        return prompt.create_instance(message_names=message_names, **variables)
+        return prompt.create_instance(input_schema=input_schema)
+
+    async def prompt_instance_from_input_schema(
+        self,
+        input_schema: RenderableModel,
+    ) -> PromptInstance:
+        """
+        Create a new PromptInstance for managing LLM interactions.
+
+        Args:
+            input_schema: Input schema to use in message rendering
+        """
+        prompt = await self.get_prompt(
+            input_schema.model_version_config["metadata"].prompt_id,
+            input_schema.model_version_config["metadata"].prompt_version_id,
+        )
+        return prompt.create_instance(input_schema=input_schema)
 
     @asynccontextmanager
     async def span(
         self,
-        name: str,
-        prompt_id: UUID,
-        prompt_version_id: UUID,
-        kind: Literal["llm", "tool", "agent"],
+        prompt_instance: PromptInstance,
+        name: str | None = None,
+        kind: Literal["llm", "tool", "agent"] = "llm",
         attributes: Optional[dict[str, Any]] = None,
     ) -> AsyncGenerator[SpanContext, None]:
         """
         Creates a new span context and sets it as the current span.
 
         Example:
-            async with client.span("agent_task", prompt_id=req_id, prompt_version_id=ver_id, kind="agent"):
+            async with client.span(name="agent_task", prompt_instance=prompt_instance, kind="agent"):
                 # Do work within the span
                 await client.log_event(llm_event)
         """
+        if name is None:
+            name = prompt_instance.base_prompt.name
         async with self.telemetry_client.span(
+            prompt_instance=prompt_instance,
             name=name,
-            prompt_id=prompt_id,
-            prompt_version_id=prompt_version_id,
             kind=kind,
             attributes=attributes,
         ) as span_context:
@@ -462,3 +550,32 @@ class MoxnClient(TelemetryTransportBackend):
             span_id: Optional specific span ID to log to (uses current span if None)
         """
         await self.telemetry_client.log_event(event=event, span_id=span_id)
+
+    async def send_telemetry_log_and_get_signed_url(
+        self, request: SignedUrlRequest
+    ) -> SignedUrlResponse:
+        """Send telemetry log and get a signed URL for uploading large data."""
+        logger.debug(f"Sending telemetry log and getting signed URL: {request}")
+
+        try:
+            # Ensure proper serialization of all fields
+            json_data = request.model_dump(
+                exclude_none=True,
+                mode="json",
+                by_alias=True,
+            )
+
+            response = await self.client.post(
+                "/telemetry/log-event-signed-url",
+                json=json_data,
+            )
+            response.raise_for_status()
+
+            return SignedUrlResponse.model_validate(response.json())
+
+        except httpx.TimeoutException as e:
+            logger.error(f"Signed URL request timed out: {e}")
+            raise RuntimeError("Telemetry signed URL request timed out") from e
+        except Exception as e:
+            logger.error(f"Signed URL request failed: {e}", exc_info=True)
+            raise RuntimeError("Failed to get signed URL for telemetry data") from e
