@@ -1,4 +1,14 @@
-from typing import Literal, overload, Any, cast, Sequence, Optional
+from typing import (
+    Literal,
+    ClassVar,
+    overload,
+    Any,
+    cast,
+    Sequence,
+    Optional,
+    runtime_checkable,
+    Protocol,
+)
 from uuid import UUID
 
 from anthropic.types import Message as AnthropicMessage
@@ -12,6 +22,29 @@ from moxn.base_models.content import AnthropicMessagesParam, OpenAIMessagesParam
 from moxn.base_models.telemetry import LLMEvent
 
 
+@runtime_checkable
+class RenderableModel(Protocol):
+    model_version_config: ClassVar[dict[str, Any]]
+
+    def model_dump(
+        self,
+        *,
+        mode: Literal["json", "python"] | str = "python",
+        include: Any = None,
+        exclude: Any = None,
+        context: dict[str, Any] | None = None,
+        by_alias: bool = False,
+        exclude_unset: bool = False,
+        exclude_defaults: bool = False,
+        exclude_none: bool = False,
+        round_trip: bool = False,
+        warnings: bool | Literal["none", "warn", "error"] = True,
+        serialize_as_any: bool = False,
+    ) -> Any: ...
+
+    def render(self, **kwargs: Any) -> Any: ...
+
+
 # Add these Pydantic models for image handling
 class ImageProviderOptions(BaseModel):
     mime: str
@@ -126,11 +159,13 @@ class PromptInstance:
         self,
         base_prompt: Prompt,
         selected_messages: list[pr.Message],
-        variables: dict | None = None,
+        input_schema: RenderableModel | None = None,
+        render_kwargs: dict[str, Any] | None = None,
     ):
         self.base_prompt = base_prompt
         self.messages = selected_messages
-        self.variables = variables or {}
+        self.input_schema = input_schema
+        self.render_kwargs = render_kwargs or {}
         self.conversation_history: list[pr.Message] = []
 
     @property
@@ -145,16 +180,22 @@ class PromptInstance:
     def from_prompt(
         cls,
         prompt: Prompt,
+        input_schema: RenderableModel | None = None,
+        render_kwargs: dict[str, Any] | None = None,
         message_names: list[str] | None = None,
         messages: list[pr.Message] | None = None,
-        **variables,
     ) -> "PromptInstance":
         """Create a PromptInstance from a base Prompt."""
         if message_names and messages:
             raise ValueError("Cannot specify both message_names and messages")
 
         selected_messages = prompt._get_selected_messages(message_names, messages)
-        return cls(prompt, selected_messages, variables)
+        return cls(
+            base_prompt=prompt,
+            input_schema=input_schema,
+            render_kwargs=render_kwargs,
+            selected_messages=selected_messages,
+        )
 
     @overload
     def append_message(
@@ -366,14 +407,23 @@ class PromptInstance:
                     all_messages[0].to_provider_message_param(
                         base_models.Provider.ANTHROPIC,
                         role=base_models.MessageRole.SYSTEM,
-                        **self.variables,
+                        **(
+                            self.input_schema.render(**self.render_kwargs)
+                            if self.input_schema
+                            else {}
+                        ),
                     )
                 ]
 
                 # Get messages from remaining messages
                 messages = [
                     p.to_provider_message_param(
-                        base_models.Provider.ANTHROPIC, **self.variables
+                        base_models.Provider.ANTHROPIC,
+                        **(
+                            self.input_schema.render(**self.render_kwargs)
+                            if self.input_schema
+                            else {}
+                        ),
                     )
                     for p in all_messages[1:]
                 ]
@@ -384,7 +434,12 @@ class PromptInstance:
                 # No system message, just regular messages
                 messages = [
                     p.to_provider_message_param(
-                        base_models.Provider.ANTHROPIC, **self.variables
+                        base_models.Provider.ANTHROPIC,
+                        **(
+                            self.input_schema.render(**self.render_kwargs)
+                            if self.input_schema
+                            else {}
+                        ),
                     )
                     for p in all_messages
                 ]
@@ -396,7 +451,12 @@ class PromptInstance:
             # For OpenAI, convert all messages to messages
             messages = [
                 p.to_provider_message_param(
-                    base_models.Provider.OPENAI, **self.variables
+                    base_models.Provider.OPENAI,
+                    **(
+                        self.input_schema.render(**self.render_kwargs)
+                        if self.input_schema
+                        else {}
+                    ),
                 )
                 for p in all_messages
             ]
@@ -446,8 +506,6 @@ class PromptInstance:
         self,
         response: AnthropicMessage,
         provider: Literal[base_models.Provider.ANTHROPIC],
-        raw_input: BaseModel | None,
-        rendered_input: dict[str, Any] | None,
         attributes: dict | None = None,
     ) -> LLMEvent: ...
 
@@ -456,8 +514,6 @@ class PromptInstance:
         self,
         response: ChatCompletion,
         provider: Literal[base_models.Provider.OPENAI],
-        raw_input: BaseModel | None,
-        rendered_input: dict[str, Any] | None,
         attributes: dict | None = None,
     ) -> LLMEvent: ...
 
@@ -465,8 +521,6 @@ class PromptInstance:
         self,
         response: ChatCompletion | AnthropicMessage,
         provider: base_models.Provider,
-        raw_input: BaseModel | None,
-        rendered_input: dict[str, Any] | None,
         attributes: dict | None = None,
     ) -> LLMEvent:
         """Creates an LLM event from the current state."""
@@ -486,7 +540,15 @@ class PromptInstance:
                 tool_call.model_dump(by_alias=True)
                 for tool_call in parsed_response.tool_calls
             ],
-            raw_input=raw_input.model_dump(by_alias=True) if raw_input else None,
-            rendered_input=rendered_input,
+            raw_input=(
+                self.input_schema.model_dump(by_alias=True)
+                if self.input_schema
+                else None
+            ),
+            rendered_input=(
+                self.input_schema.render(**self.render_kwargs)
+                if self.input_schema
+                else None
+            ),
             attributes=attributes,
         )
